import argparse
import os
import time
from ast import parse
from statistics import mean
from typing import Callable, Optional, Sequence

import GPUtil  # type: ignore

from .sentinel import Sentinel, State

TICK_TIME_S = 1


def shutdown():
    os.system("sudo shutdown -h now")


def reboot():
    os.system("sudo reboot")


def create_kill_fn(action: str) -> Callable:
    if action == "end_process":
        return exit
    elif action == "shutdown":
        return shutdown
    elif action == "reboot":
        return reboot
    else:
        raise Exception("Unknown action")


def get_gpu_usage(device_ids: Optional[Sequence[int]]) -> float:
    """Get the avg GPU usage across devices."""
    gpu_usage = []

    devices = GPUtil.getGPUs()
    for device in devices:
        if device_ids is None:
            gpu_usage.append(device.load)
        else:
            if device.id in device_ids:
                gpu_usage.append(device.load)

    return mean(gpu_usage)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="A monitor for your GPU")
    parser.add_argument(
        "--arm_duration",
        default=30,
        type=int,
        help="How many seconds of activity to wait before arming the sentinel.",
    )
    parser.add_argument(
        "--arm_threshold",
        default=0.8,
        type=float,
        help="What level of utilization is considered activity",
    )
    parser.add_argument(
        "--kill_duration",
        default=30,
        type=int,
        help="How many seconds of inactivity to wait before running the kill function.",
    )
    parser.add_argument(
        "--kill_threshold",
        default=0.7,
        type=float,
        help="What level of utilization is considered inactivity",
    )
    parser.add_argument(
        "--kill_action",
        choices=["end_process", "shutdown", "reboot"],
        default="end_process",
        help="What to do when the kill trigger is hit",
    )
    parser.add_argument(
        "--gpu_devices",
        nargs="+",
        required=False,
        help="Which GPU devices to average (empty for all)",
    )

    return parser.parse_args()


def main():
    args = parse_args()
    if args.kill_action in ["shutdown", "reboot"]:
        if os.geteuid() != 0:
            print(
                "WARNING: This program requires sudo if it is going to shutdown the computer.\n"
                "If you are have not setup your environment appropriately, the shutdown\n"
                "command will request a password and may fail to shutdown your machine.\n"
                "\n"
                "We suggest configuring your machine to not require a sudo password for\n"
                "the commands you want to run. See https://askubuntu.com/a/168885\n"
            )
    print(
        f"Starting GPU Sentinel, will arm once average GPU usage is at least {args.arm_threshold:.0%} for {args.arm_duration} seconds"
    )
    if args.gpu_devices is None:
        print(f"Monitoring all GPUs on the system")
    else:
        print(f"Monitoring GPUs {args.gpu_devices} only.")
        print("These GPUs are user specified and may not exist on the system.")

    kill_fn = create_kill_fn(args.kill_action)
    sentinel = Sentinel(
        arm_duration=args.arm_duration,
        arm_threshold=args.arm_threshold,
        kill_duration=args.kill_duration,
        kill_threshold=args.kill_threshold,
        kill_fn=kill_fn,
    )
    gpu_device_ids = None
    if args.gpu_devices:
        gpu_device_ids = [int(i) for i in args.gpu_devices]

    # Run the program until it's shutdown or we kill things.
    print("Monitoring for low GPU usage! Press ctrl+C to exit")
    show_arm_warning = True
    while True:
        gpu_usage = get_gpu_usage(device_ids=gpu_device_ids)
        sentinel.tick(gpu_usage)

        if sentinel.state == State.ARMED and show_arm_warning:
            print(
                f"GPU Sentinel is armed, will '{args.kill_action}' when values drop below {args.kill_threshold:.0%} for {args.kill_duration} seconds"
            )
            show_arm_warning = False

        time.sleep(TICK_TIME_S)


if __name__ == "__main__":
    main()



import enum
from typing import Callable


class State(enum.Enum):
    SAFE = 0
    ARMED = 1


class Sentinel:
    def __init__(
        self,
        arm_duration: int,
        arm_threshold: float,
        kill_duration: int,
        kill_threshold: float,
        kill_fn: Callable,
    ):
        """Create the GPU sentinel.

        The sentinel works by waiting to arm until a certain threshold has been reached for a set duration.
        Then it will wait until a certain threshold has not been exceeded for a set duration, and run the kill_fn
        then.

        Args:
            arm_duration: How long in ticks to wait for the threshold condition to be met before arming.
            arm_threshold: What GPU percentage required to hit threshold.
            kill_duration: How long in ticks to wait for the kill condition to be met before killing.
            kill_threshold: If the GPU drops under this percentage, consider the kill condition met.
            kill_fn: The function to call when killing, might shutdown machine, etc.
        """
        self.max_arm_duration = 60 * 60
        self.max_kill_duration = 60 * 60

        assert (
            arm_duration < self.max_arm_duration
        ), f"arm_duration can at most be {self.max_arm_duration} ticks"
        assert (
            kill_duration < self.max_kill_duration
        ), f"kill_duration can at most be {self.max_arm_duration} ticks"

        self.arm_duration = arm_duration
        self.arm_threshold = arm_threshold
        self.kill_duration = kill_duration
        self.kill_threshold = kill_threshold
        self.kill_fn = kill_fn

        self.state = State.SAFE
        self.seconds_to_transition = self.arm_duration

    def tick(self, current_value: float):
        """Run this to advance the system."""
        if self.state is State.SAFE:
            # If the new value exceeds the threshold, continue counting down.
            # Otherwise reset the timer.
            if current_value > self.arm_threshold:
                self.seconds_to_transition -= 1
            else:
                self.seconds_to_transition = self.arm_duration

            # Check if we should arm.
            if self.seconds_to_transition <= 0:
                self.state = State.ARMED
                self.seconds_to_transition = self.kill_duration
        elif self.state is State.ARMED:
            # If the new value is lower than the threshold, continue counting down.
            # Otherwise reset the timer.
            if current_value < self.kill_threshold:
                self.seconds_to_transition -= 1
            else:
                self.seconds_to_transition = self.kill_duration

            # Check if we should kill
            if self.seconds_to_transition <= 0:
                self.kill_fn()



import unittest
from unittest.mock import Mock

from .sentinel import Sentinel, State


class TestSentinel(unittest.TestCase):
    def test_arm(self):
        kill_fn = Mock()
        sentinel = Sentinel(
            arm_duration=5,
            arm_threshold=0.8,
            kill_duration=4,
            kill_threshold=0.7,
            kill_fn=kill_fn,
        )

        # When we start, state should be SAFE
        self.assertEqual(sentinel.state, State.SAFE)

        # We tick for 10 without hitting the threshold
        for _ in range(10):
            sentinel.tick(0.2)
        self.assertEqual(sentinel.state, State.SAFE)

        # Do we arm correctly?
        for _ in range(4):
            sentinel.tick(0.9)
        self.assertEqual(sentinel.state, State.SAFE)
        sentinel.tick(0.9)
        self.assertEqual(sentinel.state, State.ARMED)
        kill_fn.assert_not_called()

    def test_doesnt_arm(self):
        kill_fn = Mock()
        sentinel = Sentinel(
            arm_duration=5,
            arm_threshold=0.8,
            kill_duration=4,
            kill_threshold=0.7,
            kill_fn=kill_fn,
        )

        # When we start, state should be SAFE
        self.assertEqual(sentinel.state, State.SAFE)

        # Almost arm
        for _ in range(4):
            sentinel.tick(0.9)
        self.assertEqual(sentinel.state, State.SAFE)

        # Go back to safed
        for _ in range(4):
            sentinel.tick(0.2)
        self.assertEqual(sentinel.state, State.SAFE)

        # Go back to armed
        for _ in range(5):
            sentinel.tick(0.9)
        self.assertEqual(sentinel.state, State.ARMED)
        kill_fn.assert_not_called()

    def test_kill(self):
        kill_fn = Mock()
        sentinel = Sentinel(
            arm_duration=5,
            arm_threshold=0.8,
            kill_duration=4,
            kill_threshold=0.7,
            kill_fn=kill_fn,
        )

        # When we start, state should be SAFE
        self.assertEqual(sentinel.state, State.SAFE)

        # Arm
        for _ in range(5):
            sentinel.tick(0.9)
        self.assertEqual(sentinel.state, State.ARMED)

        # Tick for 10 without hitting
        for _ in range(10):
            sentinel.tick(0.9)
        self.assertEqual(sentinel.state, State.ARMED)

        # Tick for 4 and kill
        for _ in range(4):
            sentinel.tick(0.3)
        kill_fn.assert_called_once()


if __name__ == "__main__":
    unittest.main()
